-- Spell IDs for the Roll The Bones Buffs
local RollTheBones = 315508
local GrandMelee = "193358"
local RuthlessPrecision = "193357"
local TrueBearing = "193359"
local Broadsides = "193356"
local SkullAndCrossbones = "199603"
local BuriedTreasure = "199600"

--[[ Number of buffs active. Used to determine if we have a 1, 2 or 5 roll.
     This is calculated after every Roll The Bones use.
--]]
local amount = 0

-- EditBox to show statistics which can be copied to the clipboard.
local TextFrame = CreateFrame("Frame", "RTBS_TextFrame", UIParent, "BackdropTemplate")
TextFrame:SetSize(300, 400)
TextFrame:SetPoint("LEFT", UIParent, "LEFT", 3, 10)
TextFrame:SetFrameStrata("MEDIUM")
TextFrame:SetBackdrop({
	bgFile = [[Interface\Buttons\WHITE8x8]],
	insets = {left = 3, right = 3, top = 4, bottom = 3}
})
TextFrame:SetBackdropColor(0, 0, 0, 0.7)
TextFrame:Hide()

--Make movable
TextFrame:EnableMouse(true)
TextFrame:SetMovable(true)
TextFrame:RegisterForDrag("LeftButton")
TextFrame:SetScript("OnDragStart", TextFrame.StartMoving)
TextFrame:SetScript("OnDragStop", TextFrame.StopMovingOrSizing)

-- Create the EditBox
TextBox = CreateFrame("EditBox", "RTBS_TextBox", TextFrame)
TextBox:SetSize(300, 400)
TextBox:SetMaxLetters(99999)
TextBox:SetMultiLine(true)
TextBox:SetAutoFocus(false)
TextBox:SetFont("Fonts\\ARIALN.ttf", 14, "THINOUTLINE")
TextBox:SetScript("OnEscapePressed", function() TextBox:ClearFocus() end)

-- Create the ScrollFrame
local ScrollFrame = CreateFrame("ScrollFrame", "RTBS_TextBoxScroll", TextFrame, "UIPanelScrollFrameTemplate")
ScrollFrame:SetPoint("TOPLEFT", TextFrame, "TOPLEFT", 8, -8)
ScrollFrame:SetPoint("BOTTOMRIGHT", TextFrame, "BOTTOMRIGHT", -30, 8)
ScrollFrame:SetScrollChild(TextBox)

-- Event Frame definition
local EventFrame = CreateFrame("Frame")
EventFrame:RegisterEvent("ADDON_LOADED")

EventFrame:SetScript("OnEvent", function(self, event, ...)
	if(event == "ADDON_LOADED" and ... == "RollTheBonesStatistics") then
		-- Only load the addon if current class is rogue.
		classDisplayName, class, classId = UnitClass("player")

		if(classId == 4) then		
			-- Set event handler for talent switch.
			EventFrame:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")

			-- Initialize saved variables if called for the first time.
			RTBS_SetStandardValues()

			--[[ We need to continue with the addon init process with the following Event.
				 In ADDON_LOADED we cannot get the current spec.
			--]]
			EventFrame:RegisterEvent("PLAYER_LOGIN")
		end

	elseif(event == "PLAYER_LOGIN") then
		-- Register event if the current specialization is outlaw.
		local currentSpecIndex = GetSpecialization()
		local currentSpecId = currentSpecIndex and select(1, GetSpecializationInfo(currentSpecIndex)) or "None"

		if(currentSpecId == 260) then
			EventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
			RTBS_ConfigureDisplay()
		end

	-- Register/unregister events if the current specialization changes.
	elseif(event == "ACTIVE_TALENT_GROUP_CHANGED") then
		if(GetSpecializationInfo(GetSpecialization()) == 260) then
			EventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
			RTBS_ConfigureDisplay()
		else
			EventFrame:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
			RTBS_ConfigureDisplay()
		end
	
	--[[
		Main event of the addon. It is called each time called the player uses Roll The Bones.
		Here we check which buffs were rolled and add those to the statistics.
	--]]
	elseif(event == "UNIT_SPELLCAST_SUCCEEDED") then
		 -- Check if the player used Roll The Bones. Return if not.
		local unit, spellString, spellId = ...

		if(unit ~= "player" or spellId ~= RollTheBones) then
			return
		end

		RTBS_DB.rolls = RTBS_DB.rolls + 1
		RTBS_DB.overallRolls = RTBS_DB.overallRolls + 1

		-- Set all buffs to inactive.
		for buff in pairs(RTBS_DB.buffs) do
			RTBS_DB.buffs[buff]["active"] = false
		end

		--[[ Check how many buffs are active. Walk through all player buffs.
			 After Patch 8.0.1 we cannot use UnitBuff with a spell name any more.
			 
			 We only want to count the buffs here that come from Roll The Bones usage. Else any
			 extra buffs like those of the T21 4p set bonus will result in a wrong number of buffs calculated.
			 As of now, we regard any buff with a duration of <= 10 seconds not as an RTB buff, since using RTB always
			 creates buffs at least 12 seconds long.
		--]]
		for i = 1, 40 do
			local name, icon, count, dispelType, duration, expires, caster, isStealable, nameplateShowPersonal, spellId, canApplyAura, isBossDebuff, _, nameplateShowAll, timeMod, value1, value2, value3 = UnitBuff("player" , i)
			spellId = tostring(spellId)

			if(RTBS_DB.buffs[spellId] ~= nil and duration > 10) then
				RTBS_DB.buffs[spellId]["active"] = true
				RTBS_DB.buffs[spellId]["timesRolled"] = RTBS_DB.buffs[spellId]["timesRolled"] + 1
				amount = amount + 1
			end
		end
		
		-- Ending a streak.
		if(amount > 1) then
			-- Check if the record has to be updated.
			if(RTBS_DB.rolls > RTBS_DB.rollsRecord) then
				RTBS_DB.rollsRecord = RTBS_DB.rolls
			end

			-- Write streak statistics and add a streak ending roll.
			RTBS_DB.streakEnd[RTBS_DB.rolls] = RTBS_DB.streakEnd[RTBS_DB.rolls] + 1
			RTBS_DB.endingRolls = RTBS_DB.endingRolls + 1
			RTBS_DB.rolls = 0
		end

		-- Updating buff rolls.
		if(amount == 1) then
			RTBS_DB.num1Buffs = RTBS_DB.num1Buffs + 1
		elseif(amount == 2) then
			RTBS_DB.num2Buffs = RTBS_DB.num2Buffs + 1
		elseif(amount == 5) then
			RTBS_DB.num5Buffs = RTBS_DB.num5Buffs + 1
		end
		
		amount = 0

		-- Update the textual output.
		RTBS_UpdateTextBox()
	end
end)

--[[ This is called after every roll and updates the statistics in the
	 Text Box.
--]]
function RTBS_UpdateTextBox()
	local output = ""

	-- Textual output
	if(RTBS_DB.text == "true") then
		output = "|cffEEF56EOverall Rolls:|r " .. RTBS_DB.overallRolls .. "\n"
		output = output .. "|cffEEF56ESingle Streak Record:|r " .. RTBS_DB.rollsRecord .. "\n\n"
		output = output .. "|cffEEF56EIndividual Chances:|r\n"
		output = output .. "Grand Melee: " .. RTBS_DB.buffs[GrandMelee]["timesRolled"] .. " (" .. RTBS_Round((RTBS_DB.buffs[GrandMelee]["timesRolled"] / RTBS_DB.overallRolls * 100), RTBS_DB.roundingPrecision) .. "%)\n"
		output = output .. "Ruthless Precision: " .. RTBS_DB.buffs[RuthlessPrecision]["timesRolled"] .. " (" .. RTBS_Round((RTBS_DB.buffs[RuthlessPrecision]["timesRolled"] / RTBS_DB.overallRolls * 100), RTBS_DB.roundingPrecision) .. "%)\n"
		output = output .. "True Bearing: " .. RTBS_DB.buffs[TrueBearing]["timesRolled"] .. " (" .. RTBS_Round((RTBS_DB.buffs[TrueBearing]["timesRolled"] / RTBS_DB.overallRolls * 100), RTBS_DB.roundingPrecision) .. "%)\n"
		output = output .. "Broadsides: " .. RTBS_DB.buffs[Broadsides]["timesRolled"] .. " (" .. RTBS_Round((RTBS_DB.buffs[Broadsides]["timesRolled"] / RTBS_DB.overallRolls * 100), RTBS_DB.roundingPrecision) .. "%)\n"
		output = output .. "Skull And Crossbones: " .. RTBS_DB.buffs[SkullAndCrossbones]["timesRolled"] .. " (" .. RTBS_Round((RTBS_DB.buffs[SkullAndCrossbones]["timesRolled"] / RTBS_DB.overallRolls * 100), RTBS_DB.roundingPrecision) .. "%)\n"
		output = output .. "Buried Treasure: " .. RTBS_DB.buffs[BuriedTreasure]["timesRolled"] .. " (" .. RTBS_Round((RTBS_DB.buffs[BuriedTreasure]["timesRolled"] / RTBS_DB.overallRolls * 100), RTBS_DB.roundingPrecision) .. "%)\n\n"
		output = output .. "|cffEEF56EChances For Number Of Buffs:|r\n"
		output = output .. "1 Buff: ".. RTBS_DB.num1Buffs .. " (" .. RTBS_Round(RTBS_DB.num1Buffs / RTBS_DB.overallRolls * 100, RTBS_DB.roundingPrecision) .. "%)\n"
		output = output .. "2 Buff: ".. RTBS_DB.num2Buffs .. " (" .. RTBS_Round(RTBS_DB.num2Buffs / RTBS_DB.overallRolls * 100, RTBS_DB.roundingPrecision) .. "%)\n"
		output = output .. "5 Buff: ".. RTBS_DB.num5Buffs .. " (" .. RTBS_Round(RTBS_DB.num5Buffs / RTBS_DB.overallRolls * 100, RTBS_DB.roundingPrecision) .. "%)\n\n"
		output = output .. "|cffEEF56EStreak Ends At:|r\n"

		for i = 1, RTBS_DB.rollsRecord, 1 do
			output = output .. i .. " rolls: " .. RTBS_DB.streakEnd[i] .. " (" .. RTBS_Round(RTBS_DB.streakEnd[i] / RTBS_DB.endingRolls * 100, RTBS_DB.roundingPrecision) .. "%)\n"
		end

	-- JSON output
	elseif(RTBS_DB.json == "true") then
		output = "{\n"
		output = output .. "    \"overall_rolls\" : " .. RTBS_DB.overallRolls .. ",\n"
		output = output .. "    \"single_streak_record\" : " .. RTBS_DB.rollsRecord .. ",\n"
		output = output .. "    \"grand_melee\" : " .. RTBS_DB.buffs[GrandMelee]["timesRolled"] .. ",\n"
		output = output .. "    \"ruthless_precision\" : " .. RTBS_DB.buffs[RuthlessPrecision]["timesRolled"] .. ",\n"
		output = output .. "    \"true_bearing\" : " .. RTBS_DB.buffs[TrueBearing]["timesRolled"] .. ",\n"
		output = output .. "    \"broadsides\" : " .. RTBS_DB.buffs[Broadsides]["timesRolled"] .. ",\n"
		output = output .. "    \"skull_and_crossbones\" : " .. RTBS_DB.buffs[SkullAndCrossbones]["timesRolled"] .. ",\n"
		output = output .. "    \"buried_treasure\" : " .. RTBS_DB.buffs[BuriedTreasure]["timesRolled"] .. ",\n"
		output = output .. "    \"1_buff\" : " .. RTBS_DB.num1Buffs .. ",\n"
		output = output .. "    \"2_buff\" : " .. RTBS_DB.num2Buffs .. ",\n"
		output = output .. "    \"5_buff\" : " .. RTBS_DB.num5Buffs .. ",\n"
		output = output .. "    \"streak_ends\" : {\n"

		for i=1, RTBS_DB.rollsRecord, 1 do
			output = output .. "        \"" .. i .. "\" : " .. RTBS_DB.streakEnd[i]

			if(i < RTBS_DB.rollsRecord) then
				output = output .. ",\n"
			end
		end

		output = output .. "\n    }\n"
		output = output .. "}"
	end
	
	TextBox:SetText(output)
end

--[[ This function rounds a number to the desired decimal place.

		@param number			The number to round
		@param digits			Decimal places
		
		@return Rounded number
--]]
function RTBS_Round(number, digits)
	local shift = 10 ^ digits

	return floor(number * shift + 0.5) / shift
end

--[[ Set standard saved variables as a table into RTBS_DB. Check if each entry exists
     and if not set standard values.
--]]
function RTBS_SetStandardValues()
	if(RTBS_DB == nil) then
		RTBS_DB = {}
	end

	--[[
		Buff Table. The index "active" is written by the main event but currently not in use.
		"timesRolled" saves the total amount of times a buff was rolled.
	--]]
	if(not RTBS_DB.buffs) then
		RTBS_DB.buffs = {
			[GrandMelee] = {
				["active"] = false,
				["timesRolled"] = 0,
			},
			[RuthlessPrecision] = {
				["active"] = false,
				["timesRolled"] = 0,
			},
			[TrueBearing] = {
				["active"] = false,
				["timesRolled"] = 0,
			},
			[Broadsides] = {
				["active"] = false,
				["timesRolled"] = 0,
			},
			[SkullAndCrossbones] = {
				["active"] = false,
				["timesRolled"] = 0,
			},
			[BuriedTreasure] = {
				["active"] = false,
				["timesRolled"] = 0,
			},
		}
	end

	-- Decimal places to show in calculations.
	if(not RTBS_DB.roundingPrecision) then
		RTBS_DB.roundingPrecision = 2
	end

	--[[ Prepopulated array for saving streak ends.
		 Three single buff rolls resulting in a roll with more than one buff would add 1 to streakEnd[4].
		 Having multiple two-buff rolls back to back is a "one streak ender" resulting in +1 to streakEnd[1].
	--]]
	if(not RTBS_DB.streakEnd) then
		RTBS_DB.streakEnd = {}

		for i=1, 100 do
			RTBS_DB.streakEnd[i] = 0
		end
	end

	-- Number of rolls since the last reset.
	if(not RTBS_DB.overallRolls) then
		RTBS_DB.overallRolls = 0
	end

	-- Number of single-buff rolls in a row. Adds up every time you consecutively roll only one buff.
	if(not RTBS_DB.rolls) then
		RTBS_DB.rolls = 0
	end

	-- Number of highest single buff streak.
	if(not RTBS_DB.rollsRecord) then
		RTBS_DB.rollsRecord = 0
	end

	-- Number of rolls that ended a streak. A roll sequence of 1 > 1 > 1 > 2 > 1 > 3 > 2 has three streak ending rolls (2, 3 and 2).
	if(not RTBS_DB.endingRolls) then
		RTBS_DB.endingRolls = 0
	end

	-- Number of times 1, 2 and 5 buffs were rolled.
	if(not RTBS_DB.num1Buffs) then
		RTBS_DB.num1Buffs = 0
	end

	if(not RTBS_DB.num2Buffs) then
		RTBS_DB.num2Buffs = 0
	end

	if(not RTBS_DB.num5Buffs) then
		RTBS_DB.num5Buffs = 0
	end

	-- Whether to show the text output or not.
	if(not RTBS_DB.show) then
		RTBS_DB.show = "true"
	end

	-- Whether to show a textual or json representation of data.
	if(not RTBS_DB.text) then
		RTBS_DB.text = "true"
	end

	if(not RTBS_DB.json) then
		RTBS_DB.json = "false"
	end
	
	RTBS_ConfigureDisplay()
	RTBS_UpdateTextBox()
end

-- Show/hide the TextFrame depending on the user's settings and spec.
function RTBS_ConfigureDisplay()
	local currentSpecIndex = GetSpecialization()
	local currentSpecId = currentSpecIndex and select(1, GetSpecializationInfo(currentSpecIndex)) or "None"

	if(currentSpecId == 260) then
		if(RTBS_DB.show == "true") then
			TextFrame:Show()
		else
			TextFrame:Hide()
		end
	else
		TextFrame:Hide()
	end
end

-- Setting everything to zero and updates the text box.
function RTBS_ResetStatistics()
	RTBS_DB.buffs = {
		[RuthlessPrecision] = {
			["active"] = false,
			["timesRolled"] = 0,
		},
		[TrueBearing] = {
			["active"] = false,
			["timesRolled"] = 0,
		},
		[SkullAndCrossbones] = {
			["active"] = false,
			["timesRolled"] = 0,
		},
		[Broadsides] = {
			["active"] = false,
			["timesRolled"] = 0,
		},
		[BuriedTreasure] = {
			["active"] = false,
			["timesRolled"] = 0,
		},
		[GrandMelee] = {
			["active"] = false,
			["timesRolled"] = 0,
		},
	}

	for i=1, 100 do
		RTBS_DB.streakEnd[i] = 0
	end

	RTBS_DB.overallRolls = 0
	RTBS_DB.rolls = 0
	RTBS_DB.rollsRecord = 0
	RTBS_DB.endingRolls = 0
	RTBS_DB.num1Buffs = 0
	RTBS_DB.num2Buffs = 0
	RTBS_DB.num5Buffs = 0
	RTBS_DB.show = "true"
	RTBS_DB.text = "true"
	RTBS_DB.json = "false"

	RTBS_UpdateTextBox()
end

-- Slash commands
SLASH_RTBS1 = '/rtbs'
function SlashCmdList.RTBS(msg, editbox)
	classDisplayName, class, classId = UnitClass("player")

	if(classId == 4) then
		local command, rest = msg:match("^(%S*)%s*(.-)$");
		if(command == "show") then
			TextFrame:Show()
			RTBS_DB.show = "true"
			print("The RTBS text box is now |cffEEF56Eshown|r.")
		elseif(command == "hide") then
			TextFrame:Hide()
			RTBS_DB.show = "false"
			print("The RTBS text box is now |cffEEF56Ehidden|r.")
		elseif(command == "reset") then
			RTBS_ResetStatistics()
			print("The RTBS statistics were reset.")
		elseif(command == "json") then
			RTBS_DB.json = "true"
			RTBS_DB.text = "false"
			RTBS_UpdateTextBox()
			print("The RTBS output was set to |cffEEF56EJSON|r.")
		elseif(command == "text") then
			RTBS_DB.json = "false"
			RTBS_DB.text = "true"
			RTBS_UpdateTextBox()
			print("The RTBS output was set to |cffEEF56Etext|r.")
		elseif(command == "precision" and rest ~= nil and tonumber(rest) ~= nil) then
			-- Check if its a valid precision number.
			if(tonumber(rest) >= 0 and tonumber(rest) <= 10) then
				RTBS_DB.roundingPrecision = math.floor(rest)
				print("The RTBS rounding precision was set to |cffEEF56E" .. math.floor(rest) .. "|r.")
			else
				print("Please specify a rounding precision value between |cffEEF56E0|r and |cffEEF56E10|r.")
			end

			RTBS_UpdateTextBox()
		-- Print list of commands.
		elseif(command == "help" or command == "?") then
			DEFAULT_CHAT_FRAME:AddMessage("A list of |cffFF4500/rtbs|r commands:", 1, 1, 1)
			DEFAULT_CHAT_FRAME:AddMessage("|cffFF4500General:|r", 1, 1, 1)
			DEFAULT_CHAT_FRAME:AddMessage("    show/hide |cffEEF56E- display the textbox|r", 0.863, 0.863, 0.863)
			DEFAULT_CHAT_FRAME:AddMessage("    reset |cffEEF56E- reset all statistics|r", 0.863, 0.863, 0.863)
			DEFAULT_CHAT_FRAME:AddMessage("    json/text |cffEEF56E- switch between text and json output|r", 0.863, 0.863, 0.863)
			DEFAULT_CHAT_FRAME:AddMessage("    precision [0-10] |cffEEF56E- the number of decimal places to output|r", 0.863, 0.863, 0.863)
			DEFAULT_CHAT_FRAME:AddMessage("For a full reference check: https://mods.curse.com/addons/wow/roll-the-bones-statistics.")
		else
			print("This is not a command of Roll The Bones Statistics. Please consider reading the documentation at https://mods.curse.com/addons/wow/roll-the-bones-statistics or type |cffFF4500/rtbs help|r.")
		end
	else
		print("You need to be on your |cffffff00Rogue|r to change the settings of Roll The Bones Statistics.")
	end
end